@@ -11,6 +11,7 @@
 */
 
 #include <iostream>
+#include <chrono>
 #include <list>
 
 #include <zypp/base/Logger.h>
@@ -42,19 +43,129 @@ using std::endl;
 
 namespace internal {
   using namespace zypp;
+  /// \brief Bottleneck filtering all DownloadProgressReport issued from Media[Muli]Curl.
+  /// - Optional files will send no report until data are actually received (we know it exists).
+  /// - Control the progress report frequency passed along to the application.
+  struct OptionalDownloadProgressReport : public callback::ReceiveReport<media::DownloadProgressReport>
+  {
+    using TimePoint = std::chrono::steady_clock::time_point;
+
+    OptionalDownloadProgressReport( bool isOptional=false )
+    : _oldRec { Distributor::instance().getReceiver() }
+    , _isOptional { isOptional }
+    { connect(); }
+
+    ~OptionalDownloadProgressReport()
+    { if ( _oldRec ) Distributor::instance().setReceiver( *_oldRec ); else Distributor::instance().noReceiver(); }
+
+    void reportbegin() override
+    { if ( _oldRec ) _oldRec->reportbegin(); }
+
+    void reportend() override
+    { if ( _oldRec ) _oldRec->reportend(); }
+
+    void report( const UserData & userData_r = UserData() ) override
+    { if ( _oldRec ) _oldRec->report( userData_r ); }
+
+
+    void start( const Url & file_r, Pathname localfile_r ) override
+    {
+      if ( not _oldRec ) return;
+      if ( _isOptional ) {
+        // delay start until first data are received.
+        _startFile      = file_r;
+        _startLocalfile = std::move(localfile_r);
+        return;
+      }
+      _oldRec->start( file_r, localfile_r );
+    }
+
+    bool progress( int value_r, const Url & file_r, double dbps_avg_r = -1, double dbps_current_r = -1 ) override
+    {
+      if ( not _oldRec ) return true;
+      if ( notStarted() ) {
+        if ( not ( value_r || dbps_avg_r || dbps_current_r ) )
+          return true;
+        sendStart();
+      }
+
+      //static constexpr std::chrono::milliseconds minfequency { 1000 }; only needed if we'd avoid sending reports without change
+      static constexpr std::chrono::milliseconds maxfequency { 100 };
+      TimePoint now { TimePoint::clock::now() };
+      TimePoint::duration elapsed { now - _lastProgressSent };
+      if ( elapsed < maxfequency )
+        return true;  // continue
+      _lastProgressSent = now;
+      return _oldRec->progress( value_r, file_r, dbps_avg_r, dbps_current_r );
+    }
+
+    Action problem( const Url & file_r, Error error_r, const std::string & description_r ) override
+    {
+      if ( not _oldRec || notStarted() ) return ABORT;
+      return _oldRec->problem( file_r, error_r, description_r );
+    }
+
+    void finish( const Url & file_r, Error error_r, const std::string & reason_r ) override
+    {
+      if ( not _oldRec || notStarted() ) return;
+      _oldRec->finish( file_r, error_r, reason_r );
+    }
+
+  private:
+    // _isOptional also indicates the delayed start
+    bool notStarted() const
+    { return _isOptional; }
+
+    void sendStart()
+    {
+      if ( _isOptional ) {
+        // we know _oldRec is valid...
+        _oldRec->start( std::move(_startFile), std::move(_startLocalfile) );
+        _isOptional = false;
+      }
+    }
+
+  private:
+    Receiver *const _oldRec;
+    bool     _isOptional;
+    Url      _startFile;
+    Pathname _startLocalfile;
+    TimePoint _lastProgressSent;
+  };
+
   struct ProgressData
   {
-    ProgressData( CURL *_curl, time_t _timeout = 0, const zypp::Url & _url = zypp::Url(),
-      zypp::ByteCount expectedFileSize_r = 0,
-      zypp::callback::SendReport<zypp::media::DownloadProgressReport> *_report = nullptr );
+    ProgressData( CURL *curl, time_t timeout = 0, const zypp::Url & url = zypp::Url(),
+                  zypp::ByteCount expectedFileSize_r = 0,
+                  zypp::callback::SendReport<zypp::media::DownloadProgressReport> *_report = nullptr );
 
-    CURL	*curl;
-    zypp::Url	url;
-    time_t	timeout;
-    bool	reached;
-    bool      fileSizeExceeded;
+    void updateStats( double dltotal = 0.0, double dlnow = 0.0 );
+
+    int reportProgress() const;
+
+    CURL * curl()
+    { return _curl; }
+
+    bool timeoutReached() const
+    { return _timeoutReached; }
+
+    bool fileSizeExceeded() const
+    { return _fileSizeExceeded; }
+
+    ByteCount expectedFileSize() const
+    { return _expectedFileSize; }
+
+    void expectedFileSize( ByteCount newval_r )
+    { _expectedFileSize = newval_r; }
+
+  private:
+    CURL *      _curl;
+    zypp::Url	_url;
+    time_t	_timeout;
+    bool	_timeoutReached;
+    bool        _fileSizeExceeded;
+    ByteCount   _expectedFileSize;
     zypp::callback::SendReport<zypp::media::DownloadProgressReport> *report;
-    zypp::ByteCount _expectedFileSize;
 
     time_t _timeStart	= 0;	///< Start total stats
     time_t _timeLast	= 0;	///< Start last period(~1sec)
@@ -69,38 +180,18 @@ namespace internal {
 
     double _drateTotal= 0.0;	///< Download rate so far
     double _drateLast	= 0.0;	///< Download rate in last period
-
-    void updateStats( double dltotal = 0.0, double dlnow = 0.0 );
-
-    int reportProgress() const;
-
-
-    // download rate of the last period (cca 1 sec)
-    double                                        drate_period;
-    // bytes downloaded at the start of the last period
-    double                                        dload_period;
-    // seconds from the start of the download
-    long                                          secs;
-    // average download rate
-    double                                        drate_avg;
-    // last time the progress was reported
-    time_t                                        ltime;
-    // bytes downloaded at the moment the progress was last reported
-    double                                        dload;
-    // bytes uploaded at the moment the progress was last reported
-    double                                        uload;
   };
 
 
 
-  ProgressData::ProgressData(CURL *_curl, time_t _timeout, const Url &_url, ByteCount expectedFileSize_r, zypp::callback::SendReport< zypp::media::DownloadProgressReport> *_report)
-    : curl( _curl )
-    , url( _url )
-    , timeout( _timeout )
-    , reached( false )
-    , fileSizeExceeded ( false )
-    , report( _report )
+  ProgressData::ProgressData(CURL *curl, time_t timeout, const Url &url, ByteCount expectedFileSize_r, zypp::callback::SendReport< zypp::media::DownloadProgressReport> *_report)
+    : _curl( curl )
+    , _url( url )
+    , _timeout( timeout )
+    , _timeoutReached( false )
+    , _fileSizeExceeded ( false )
     , _expectedFileSize( expectedFileSize_r )
+    , report( _report )
   {}
 
   void ProgressData::updateStats(double dltotal, double dlnow)
@@ -122,11 +213,11 @@ namespace internal {
       _timeStart = _timeLast = _timeRcv = now;
 
     // timeout condition
-    if ( timeout )
-      reached = ( (now - _timeRcv) > timeout );
+    if ( _timeout )
+      _timeoutReached = ( (now - _timeRcv) > _timeout );
 
     // check if the downloaded data is already bigger than what we expected
-    fileSizeExceeded = _expectedFileSize > 0 && _expectedFileSize < static_cast<ByteCount::SizeType>(_dnlNow);
+    _fileSizeExceeded = _expectedFileSize > 0 && _expectedFileSize < static_cast<ByteCount::SizeType>(_dnlNow);
 
     // percentage:
     if ( _dnlTotal )
@@ -148,11 +239,11 @@ namespace internal {
 
   int ProgressData::reportProgress() const
   {
-    if ( fileSizeExceeded )
+    if ( _fileSizeExceeded )
       return 1;
-    if ( reached )
+    if ( _timeoutReached )
       return 1;	// no-data timeout
-    if ( report && !(*report)->progress( _dnlPercent, url, _drateTotal, _drateLast ) )
+    if ( report && !(*report)->progress( _dnlPercent, _url, _drateTotal, _drateLast ) )
       return 1;	// user requested abort
     return 0;
   }
@@ -200,6 +291,20 @@ namespace internal {
       );
     return _value.c_str();
   }
+
+  /// Attempt to work around certain issues by autoretry in MediaCurl::getFileCopy
+  /// E.g. curl error: 92: HTTP/2 PROTOCOL_ERROR as in bsc#1205843, zypper/issues/457,...
+  class MediaCurlExceptionMayRetryInternaly : public media::MediaCurlException
+  {
+  public:
+    MediaCurlExceptionMayRetryInternaly( const Url & url_r,
+                                         const std::string & err_r,
+                                         const std::string & msg_r )
+    : media::MediaCurlException( url_r, err_r, msg_r )
+    {}
+    //~MediaCurlExceptionMayRetryInternaly() noexcept {}
+  };
+
 }
 
 
@@ -235,7 +340,6 @@ MediaCurl::MediaCurl( const Url &      url_r,
       _customHeaders(0L)
 {
   _curlError[0] = '\0';
-  _curlDebug = 0L;
 
   MIL << "MediaCurl::MediaCurl(" << url_r << ", " << attach_point_hint_r << ")" << endl;
 
@@ -301,15 +405,7 @@ void MediaCurl::checkProtocol(const Url &url) const
 
 void MediaCurl::setupEasy()
 {
-  {
-    _curlDebug = env::ZYPP_MEDIA_CURL_DEBUG();
-    if( _curlDebug > 0)
-    {
-      curl_easy_setopt( _curl, CURLOPT_VERBOSE, 1L);
-      curl_easy_setopt( _curl, CURLOPT_DEBUGFUNCTION, log_curl);
-      curl_easy_setopt( _curl, CURLOPT_DEBUGDATA, &_curlDebug);
-    }
-  }
+  ::internal::setupZYPP_MEDIA_CURL_DEBUG( _curl );
 
   curl_easy_setopt(_curl, CURLOPT_HEADERFUNCTION, log_redirects_curl);
   curl_easy_setopt(_curl, CURLOPT_HEADERDATA, &_lastRedirect);
@@ -333,9 +429,6 @@ void MediaCurl::setupEasy()
   }
   vol_settings.addHeader("Pragma:");
 
-  _settings.setTimeout(ZConfig::instance().download_transfer_timeout());
-  _settings.setConnectTimeout(CONNECT_TIMEOUT);
-
   _settings.setUserAgentString(agentString());
 
   // fill some settings from url query parameters
@@ -626,20 +719,6 @@ void MediaCurl::releaseFrom( const std::string & ejectDev )
   disconnect();
 }
 
-Url MediaCurl::getFileUrl( const Pathname & filename_r ) const
-{
-  // Simply extend the URLs pathname. An 'absolute' URL path
-  // is achieved by encoding the leading '/' in an URL path:
-  //   URL: ftp://user@server		-> ~user
-  //   URL: ftp://user@server/		-> ~user
-  //   URL: ftp://user@server//		-> ~user
-  //   URL: ftp://user@server/%2F	-> /
-  //                         ^- this '/' is just a separator
-  Url newurl( _url );
-  newurl.setPathName( ( Pathname("./"+_url.getPathName()) / filename_r ).asString().substr(1) );
-  return newurl;
-}
-
 ///////////////////////////////////////////////////////////////////
 
 void MediaCurl::getFile( const OnMediaLocation &file ) const
@@ -656,33 +735,47 @@ void MediaCurl::getFileCopy( const OnMediaLocation & srcFile , const Pathname &
 
   const auto &filename = srcFile.filename();
 
+  // Optional files will send no report until data are actually received (we know it exists).
+  OptionalDownloadProgressReport reportfilter( srcFile.optional() );
   callback::SendReport<DownloadProgressReport> report;
 
   Url fileurl(getFileUrl(filename));
 
-  bool retry = false;
+  bool firstAuth = true;  // bsc#1210870: authenticate must not return stored credentials more than once.
+  unsigned internalTry = 0;
+  static constexpr unsigned maxInternalTry = 3;
 
   do
   {
     try
     {
       doGetFileCopy( srcFile, target, report );
-      retry = false;
+      break;  // success!
     }
     // retry with proper authentication data
     catch (MediaUnauthorizedException & ex_r)
     {
-      if(authenticate(ex_r.hint(), !retry))
-        retry = true;
-      else
-      {
-        report->finish(fileurl, zypp::media::DownloadProgressReport::ACCESS_DENIED, ex_r.asUserHistory());
-        ZYPP_RETHROW(ex_r);
+      if ( authenticate(ex_r.hint(), firstAuth) ) {
+        firstAuth = false;  // must not return stored credentials again
+        continue; // retry
       }
+
+      report->finish(fileurl, zypp::media::DownloadProgressReport::ACCESS_DENIED, ex_r.asUserHistory());
+      ZYPP_RETHROW(ex_r);
     }
     // unexpected exception
     catch (MediaException & excpt_r)
     {
+      if ( typeid(excpt_r) == typeid( MediaCurlExceptionMayRetryInternaly ) ) {
+        ++internalTry;
+        if ( internalTry < maxInternalTry ) {
+          // just report (NO_ERROR); no interactive request to the user
+          report->problem(fileurl, media::DownloadProgressReport::NO_ERROR, excpt_r.asUserHistory()+_("Will try again..."));
+          continue; // retry
+        }
+        excpt_r.addHistory( str::Format(_("Giving up after %1% attempts.")) % maxInternalTry );
+      }
+
       media::DownloadProgressReport::Error reason = media::DownloadProgressReport::ERROR;
       if( typeid(excpt_r) == typeid( media::MediaFileNotFoundException )  ||
           typeid(excpt_r) == typeid( media::MediaNotAFileException ) )
@@ -693,8 +786,7 @@ void MediaCurl::getFileCopy( const OnMediaLocation & srcFile , const Pathname &
       ZYPP_RETHROW(excpt_r);
     }
   }
-  while (retry);
-
+  while ( true );
   report->finish(fileurl, zypp::media::DownloadProgressReport::NO_ERROR, "");
 }
 
@@ -854,7 +946,14 @@ void MediaCurl::evaluateCurlCode(const Pathname &filename,
           err = "User abort";
         }
         break;
-      case CURLE_SSL_PEER_CERTIFICATE:
+
+      // Attempt to work around certain issues by autoretry in MediaCurl::getFileCopy
+      case CURLE_HTTP2:
+      case CURLE_HTTP2_STREAM:
+        err = "Curl error " + str::numstring( code );
+        ZYPP_THROW(MediaCurlExceptionMayRetryInternaly(url, err, _curlError));
+        break;
+
       default:
         err = "Curl error " + str::numstring( code );
         break;
@@ -907,94 +1006,58 @@ bool MediaCurl::doGetDoesFileExist( const Pathname & filename ) const
     ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
   }
 
-  // instead of returning no data with NOBODY, we return
+  // If no head requests allowed (?head_requests=no):
+  // Instead of returning no data with NOBODY, we return
   // little data, that works with broken servers, and
   // works for ftp as well, because retrieving only headers
   // ftp will return always OK code ?
   // See http://curl.haxx.se/docs/knownbugs.html #58
-  if (  (_url.getScheme() == "http" ||  _url.getScheme() == "https") &&
-        _settings.headRequestsAllowed() )
-    ret = curl_easy_setopt( _curl, CURLOPT_NOBODY, 1L );
-  else
-    ret = curl_easy_setopt( _curl, CURLOPT_RANGE, "0-1" );
+  /// RAII Handler for temp. setting a head/range request
+  struct TempSetHeadRequest
+  {
+    TempSetHeadRequest( CURL * curl_r, bool doHttpHeadRequest_r )
+    : _curl { curl_r }
+    , _doHttpHeadRequest { doHttpHeadRequest_r }
+    {
+      if ( _doHttpHeadRequest ) {
+        curl_easy_setopt( _curl, CURLOPT_NOBODY, 1L );
+      } else {
+        curl_easy_setopt( _curl, CURLOPT_RANGE, "0-1" );
+      }
+    }
+    ~TempSetHeadRequest() {
+      if ( _doHttpHeadRequest ) {
+        curl_easy_setopt( _curl, CURLOPT_NOBODY, 0L);
+        /* yes, this is why we never got to get NOBODY working before,
+         because setting it changes this option too, and we also*
+         need to reset it
+         See: http://curl.haxx.se/mail/archive-2005-07/0073.html
+         */
+        curl_easy_setopt( _curl, CURLOPT_HTTPGET, 1L );
+      } else {
+        curl_easy_setopt( _curl, CURLOPT_RANGE, NULL );
+      }
+    }
+  private:
+    CURL * _curl;
+    bool   _doHttpHeadRequest;
+  } _guard( _curl, (_url.getScheme() == "http" || _url.getScheme() == "https") && _settings.headRequestsAllowed() );
 
-  if ( ret != 0 ) {
-    curl_easy_setopt( _curl, CURLOPT_NOBODY, 0L);
-    curl_easy_setopt( _curl, CURLOPT_RANGE, NULL );
-    /* yes, this is why we never got to get NOBODY working before,
-       because setting it changes this option too, and we also
-       need to reset it
-       See: http://curl.haxx.se/mail/archive-2005-07/0073.html
-    */
-    curl_easy_setopt( _curl, CURLOPT_HTTPGET, 1L );
-    ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
-  }
 
   AutoFILE file { ::fopen( "/dev/null", "w" ) };
   if ( !file ) {
       ERR << "fopen failed for /dev/null" << endl;
-      curl_easy_setopt( _curl, CURLOPT_NOBODY, 0L);
-      curl_easy_setopt( _curl, CURLOPT_RANGE, NULL );
-      /* yes, this is why we never got to get NOBODY working before,
-       because setting it changes this option too, and we also
-       need to reset it
-       See: http://curl.haxx.se/mail/archive-2005-07/0073.html
-      */
-      curl_easy_setopt( _curl, CURLOPT_HTTPGET, 1L );
-      if ( ret != 0 ) {
-          ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
-      }
       ZYPP_THROW(MediaWriteException("/dev/null"));
   }
 
   ret = curl_easy_setopt( _curl, CURLOPT_WRITEDATA, (*file) );
   if ( ret != 0 ) {
-      std::string err( _curlError);
-      curl_easy_setopt( _curl, CURLOPT_RANGE, NULL );
-      curl_easy_setopt( _curl, CURLOPT_NOBODY, 0L);
-      /* yes, this is why we never got to get NOBODY working before,
-       because setting it changes this option too, and we also
-       need to reset it
-       See: http://curl.haxx.se/mail/archive-2005-07/0073.html
-      */
-      curl_easy_setopt( _curl, CURLOPT_HTTPGET, 1L );
-      if ( ret != 0 ) {
-          ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
-      }
-      ZYPP_THROW(MediaCurlSetOptException(url, err));
+    ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
   }
 
   CURLcode ok = curl_easy_perform( _curl );
   MIL << "perform code: " << ok << " [ " << curl_easy_strerror(ok) << " ]" << endl;
 
-  // reset curl settings
-  if (  _url.getScheme() == "http" ||  _url.getScheme() == "https" )
-  {
-    curl_easy_setopt( _curl, CURLOPT_NOBODY, 0L);
-    if ( ret != 0 ) {
-      ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
-    }
-
-    /* yes, this is why we never got to get NOBODY working before,
-       because setting it changes this option too, and we also
-       need to reset it
-       See: http://curl.haxx.se/mail/archive-2005-07/0073.html
-    */
-    curl_easy_setopt( _curl, CURLOPT_HTTPGET, 1L);
-    if ( ret != 0 ) {
-      ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
-    }
-
-  }
-  else
-  {
-    // for FTP we set different options
-    curl_easy_setopt( _curl, CURLOPT_RANGE, NULL);
-    if ( ret != 0 ) {
-      ZYPP_THROW(MediaCurlSetOptException(url, _curlError));
-    }
-  }
-
   // as we are not having user interaction, the user can't cancel
   // the file existence checking, a callback or timeout return code
   // will be always a timeout.
@@ -1015,49 +1078,6 @@ bool MediaCurl::doGetDoesFileExist( const Pathname & filename ) const
 
 ///////////////////////////////////////////////////////////////////
 
-
-#if DETECT_DIR_INDEX
-bool MediaCurl::detectDirIndex() const
-{
-  if(_url.getScheme() != "http" && _url.getScheme() != "https")
-    return false;
-  //
-  // try to check the effective url and set the not_a_file flag
-  // if the url path ends with a "/", what usually means, that
-  // we've received a directory index (index.html content).
-  //
-  // Note: This may be dangerous and break file retrieving in
-  //       case of some server redirections ... ?
-  //
-  bool      not_a_file = false;
-  char     *ptr = NULL;
-  CURLcode  ret = curl_easy_getinfo( _curl,
-                                     CURLINFO_EFFECTIVE_URL,
-                                     &ptr);
-  if ( ret == CURLE_OK && ptr != NULL)
-  {
-    try
-    {
-      Url         eurl( ptr);
-      std::string path( eurl.getPathName());
-      if( !path.empty() && path != "/" && *path.rbegin() == '/')
-      {
-        DBG << "Effective url ("
-            << eurl
-            << ") seems to provide the index of a directory"
-            << endl;
-        not_a_file = true;
-      }
-    }
-    catch( ... )
-    {}
-  }
-  return not_a_file;
-}
-#endif
-
-///////////////////////////////////////////////////////////////////
-
 void MediaCurl::doGetFileCopy( const OnMediaLocation &srcFile , const Pathname & target, callback::SendReport<DownloadProgressReport> & report, RequestOptions options ) const
 {
     Pathname dest = target.absolutename();
@@ -1254,23 +1274,16 @@ void MediaCurl::doGetFileCopyFile( const OnMediaLocation & srcFile, const Pathna
       // otherwise it would be a user cancel
       try {
 
-        if ( progressData.fileSizeExceeded )
-          ZYPP_THROW(MediaFileSizeExceededException(url, progressData._expectedFileSize));
+        if ( progressData.fileSizeExceeded() )
+          ZYPP_THROW(MediaFileSizeExceededException(url, progressData.expectedFileSize()));
 
-        evaluateCurlCode( srcFile.filename(), ret, progressData.reached );
+        evaluateCurlCode( srcFile.filename(), ret, progressData.timeoutReached() );
       }
       catch ( const MediaException &e ) {
         // some error, we are not sure about file existence, rethrw
         ZYPP_RETHROW(e);
       }
     }
-
-#if DETECT_DIR_INDEX
-    if (!ret && detectDirIndex())
-      {
-        ZYPP_THROW(MediaNotAFileException(_url, filename));
-      }
-#endif // DETECT_DIR_INDEX
 }
 
 ///////////////////////////////////////////////////////////////////
@@ -1345,7 +1358,7 @@ int MediaCurl::progressCallback( void *clientp, double dltotal, double dlnow, do
   {
     // work around curl bug that gives us old data
     long httpReturnCode = 0;
-    if ( curl_easy_getinfo( pdata->curl, CURLINFO_RESPONSE_CODE, &httpReturnCode ) != CURLE_OK || httpReturnCode == 0 )
+    if ( curl_easy_getinfo( pdata->curl(), CURLINFO_RESPONSE_CODE, &httpReturnCode ) != CURLE_OK || httpReturnCode == 0 )
       return aliveCallback( clientp, dltotal, dlnow, ultotal, ulnow );
 
     pdata->updateStats( dltotal, dlnow );
@@ -1357,7 +1370,7 @@ int MediaCurl::progressCallback( void *clientp, double dltotal, double dlnow, do
 CURL *MediaCurl::progressCallback_getcurl( void *clientp )
 {
   internal::ProgressData *pdata = reinterpret_cast<internal::ProgressData *>(clientp);
-  return pdata ? pdata->curl : 0;
+  return pdata ? pdata->curl() : 0;
 }
 
 ///////////////////////////////////////////////////////////////////
@@ -1385,7 +1398,7 @@ void MediaCurl::resetExpectedFileSize(void *clientp, const ByteCount &expectedFi
 {
   internal::ProgressData *data = reinterpret_cast<internal::ProgressData *>(clientp);
   if ( data ) {
-    data->_expectedFileSize = expectedFileSize;
+    data->expectedFileSize( expectedFileSize );
   }
 }
 
